/*
 * Diese Datei ist Teil des CAT-Projekts.
 *
 * CAT ist Freeware; es darf unter den Bedingungen der
 * CAT Public License weitergegeben und/oder modifiziert
 * werden.
 *
 * Eine Kopie der CAT Public License liegt diesem Paket bei
 * (siehe CATPL.TXT).
 *
 * Copyright (c) 1997, Dirk Steins
 * Jegliche Veraenderung dieses Hinweises ist NICHT ERLAUBT.
 */

/*
 * Die vorliegende Version wurde nach C portiert von
 * Thorsten Otto (Thorsten_Otto@do.maus.ruhr.de)
 */



#include <infofile.h>
#include <stddef.h>
#include <mtpaths.h>
#include <catgloba.h>
#include <mte.h>
#include <datasys.h>
#include <catfiles.h>
#include <convertd.h>
#include <groupsel.h>
#include <protokol.h>
#include <confvars.h>
#include <varnamel.h>
#include <userinfo.h>

#include <mtalerts.h>
#include <mtutils.h>
#include <mtpopups.h>
#include <mtappl.h>
#include <mttextfi.h>
#include <tos.h>

#include <ro_mem.h>
#include <routinec.h>
#include <routines.h>
#include <s_endian.h>
#include <file_io.h>
#include <strconv.h>
#include <mtrsc.h>


/* Beschreibung des Datenformates:
 * Die Infofiles im Speicher werden als verkettete Liste
 * von infEntrys dargestellt. Dabei wird fuer einen Eintrag
 * nur so viel Speicher alloziert, wie unbedingt notwendig.
 * Auf Platte werden die Files als Array [0..#Infofiles]
 * von infEntrys abgespeichert, wobei diese natuerlich den
 * vollen Platz belegen.
 */

#define CREATE_DEFAULT 0


/* Magics fuer Header */
#define infCatMagic     0x43415449L  /* "CATI"                    */
#define oldInfVersion   1            /* alte Versionsnummer der Infofile.i */
#define oldInfVersionMagic 0x0140    /* altes Versionsmagic       */

#define infVersion      2            /* Versionsnummer der Infofile.i */
#define infVersionMagic 0x0100       /* Versionsmagic             */

#define maxInf          1024

typedef struct {
	CRC16 crc;
	_BOOL new;
	_BOOL avail;
	_BOOL isMaus;
	_BOOL ordered;
	_WORD orderIntervall; /* -1: Einmalig, 0: Immer, 1: taeglich */
	_ULONG lastOrdered; /* Datum in Tagen seit dem 1.1.1900 */
	_BOOL selected;
	_WORD /* tCrcState */ hasCrc;
	_UWORD res;
	_UBYTE mausShort[10];
	_UBYTE mausName[36];
	_UBYTE filename[256];
} PACKED oldInfEntry;

GLOBAL List infList;

LOCAL CONST _UBYTE *InfoGroupNames[] =
{
	"Standard-Infodateien"      ,	/* I_INFOFILE 0 */
	"Jedesmal-Infodateien"      ,	/* J_INFOFILE 1 */
	"OUT-Bl\224cke"      	    ,	/* O_INFOFILE 2 */
	"Sonstige Infodateien"      ,	/* S_INFOFILE 3 */
	"Userdefinierte Infodateien"	/* U_INFOFILE 4 */
};


/******************************************************************************/
/*** ---------------------------------------------------------------------- ***/
/******************************************************************************/

LOCAL _BOOL grComp(CONST _VOID *a1, CONST _VOID *a2, _VOID *private)
{
	CONST infEntry *p1 = *((CONST infEntry * CONST *)a1);
	CONST infEntry *p2 = *((CONST infEntry * CONST *)a2);

	UNUSED(private);
	if ((p1->i.res & ITI_GROUP_MASK) == (p2->i.res & ITI_GROUP_MASK))	
	{
		if (p1->i.isMaus == p2->i.isMaus)
			return strCaseCmp(p1->i.mausShort, p2->i.mausShort) < 0;
		return p1->i.isMaus;
	}
	return (p1->i.res & ITI_GROUP_MASK) < (p2->i.res & ITI_GROUP_MASK);
}

/*** ---------------------------------------------------------------------- ***/

/* Entfernt die Titeleintraege (MH) */
LOCAL _VOID deleteTitles(List *l)
{
	infEntry *pInf;
	
	Lists_ResetList(l);
	while ((pInf = Lists_NextEntry(l)) != NULL)
	{
		if (pInf->i.res & ITI_TITLE)
		{
			Lists_RemoveEntry(l, mem_free);
		} else
		{
			pInf->i.res |= ITI_GROUP_VISIBLE;	/* Alles sichtbar machen */
			pInf->i.res |= ITI_ENTRY_VISIBLE;
		}
	}
}

/*** ---------------------------------------------------------------------- ***/

/* sortiert die Gruppenliste und fuegt dabei die Gruppentitel ein (MH) */
GLOBAL _VOID Infofiles_SortList(List *l)
{
	_WORD lastGroup, newGroup;
	_UBYTE var_name[32];
	_BOOL visible;
	infEntry *pInf;
	infEntry *pTitle;
	
	/* Die Liste aufraeumen ... */
	deleteTitles(l);	
	
	/* ... sortieren ... */
	GroupSelect_SortAnyList(l, grComp);

	/* ... und Ueberschriften wieder einfuegen */
	Lists_ResetList(l);
	lastGroup = -1;
	visible = TRUE;
	while ((pInf = Lists_NextEntry(l)) != NULL)
	{
		/* Gruppenwechsel? */
		newGroup = (pInf->i.res & ITI_GROUP_MASK) >> ITI_GROUP_SHIFT;
		if (newGroup != lastGroup)
		{
			lastGroup = newGroup;
			
			/* Variable holen */
			make_varname(var_name, sizeof(var_name), cInfoGroupVisible, lastGroup);
			ConfVars_GetConfDefBool(var_name, &visible, TRUE);

			/* Titeleintrag erzeugen */
			pTitle = NEW(infEntry, "WriteInfTitle");
			if (pTitle == NULL)
				return;	
			MemSetZeroStruct(*pTitle);
			pTitle->i.res = lastGroup << ITI_GROUP_SHIFT;
			pTitle->i.res |= ITI_TITLE;
			if (visible)							/* Je nach Einstellung */
				pTitle->i.res |= ITI_GROUP_VISIBLE;
			pTitle->i.res |= ITI_ENTRY_VISIBLE;		/* Auf jeden Fall sichtbar */
			strBcpy(pTitle->i.mausName, InfoGroupNames[lastGroup]);

			pInf = Lists_PrevEntry(l);	/* davor eintragen */
			Lists_InsertEntry(l, pTitle);
			pInf = Lists_NextEntry(l);
		}	

		pInf->i.res &= ~ITI_GROUP_VISIBLE;		/* Wird eh nicht ausgewertet ... */
		if (visible)							/* Je nach Einstellung */
			pInf->i.res |= ITI_ENTRY_VISIBLE;
		else
			pInf->i.res &= ~ITI_ENTRY_VISIBLE;
	}	
}

/*** ---------------------------------------------------------------------- ***/

/* Liefert die interne Gruppennummer eines Infofiles (MH) */

GLOBAL _WORD Infofiles_GetInfoGroupNumber(infEntry *info)
{
	if (info->i.isMaus)
	{
		switch (info->i.mausShort[0])
		{
			case 'I': return ITI_INFOFILE_I;
			case 'Y': return ITI_INFOFILE_Y;
			case 'J': return ITI_INFOFILE_J;
			case 'O': if (strLen(info->i.mausShort) == 7 &&
						     info->i.mausShort[1] == 'U' &&
		    		         info->i.mausShort[2] == 'T')
		    		  return ITI_INFOFILE_O;
		    default:  return ITI_INFOFILE_S;
		}
	} else
	{
		return ITI_INFOFILE_U;
	}
}	

/*** ---------------------------------------------------------------------- ***/


GLOBAL _BOOL Infofiles_IsNeededInfo(CONST _UBYTE *str)
{
	/*
	 * wer das rauspatched ist selber schuld.
	 * Ausserdem tauchen die Strings auch noch an anderen
	 * Stellen auf.
	 */
	if (maus_short_name_equal(str, ITG_SHORT_NAME) ||
		maus_short_name_equal(str, YGN_SHORT_NAME) ||
		maus_short_name_equal(str, ITI_SHORT_NAME) ||
		maus_short_name_equal(str, ITK_SHORT_NAME))
		return TRUE;
	return FALSE;
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _VOID ConvertInfoV1toV2(CONST oldInfEntry *old, infEntry *new)
{
	new->i.crc = n2hs(old->crc);
	new->i.new = n2hs(old->new);
	new->i.avail = n2hs(old->avail);
	new->i.isMaus = n2hs(old->isMaus);
	new->i.ordered = n2hs(old->ordered);
	new->i.orderIntervall = n2hs(old->orderIntervall);
	new->i.lastOrdered = n2hl(old->lastOrdered);
	new->i.selected = n2hs(old->selected);
	new->i.hasCrc = n2hs(old->hasCrc);
	new->i.res    = n2hs(old->res);
	strBcpy(new->i.mausShort, old->mausShort);
	strBcpy(new->i.mausName, old->mausName);
	strBcpy(new->i.filename, old->filename);
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _BOOL clearNewProc(_VOID *entry, _VOID *info)
{
	infEntry *e = entry;

	UNUSED(info);
	e->i.new = FALSE;
	return FALSE;
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _BOOL Infofiles_Init(_VOID)
{
	if (infList.root == NULL)
		if (!Lists_CreateList(&infList, "Infofiles_Init"))
			return FALSE;
	return TRUE;
}

/*** ---------------------------------------------------------------------- ***/

	/*DYJ 5.21b */
GLOBAL _BOOL Infofiles_LoadInfoInf(_BOOL dbChanged, _BOOL clearNew)
{
	FILENO handle;
	size_t bSize;
	_ULONG size;
	_UWORD count;
	infEntry *entry;
	infEntry *infArray;
	oldInfEntry *oldInfArray;
	_UWORD i;
	FileHeaderType dbHdr;
	_BOOL isOldInf;
	_UBYTE fname[PATH_MAX];
	F_ERROR error;
	
	if (!Infofiles_Init())
		return FALSE;

	strBcpy(fname, DataPath);
	F_Path_Append(fname, infoInfFile);
	handle = CatFiles_OpenFile(fname, CatFiles_readFile);
	if (handle >= 0)
	{
		/* Jetzt erst mal Header einlesen */
		if ((error = CatFiles_ReadDbHeader(handle, &dbHdr)) != NoError)
		{
			F_ErrorAlert(error);
			CatFiles_CloseFile(handle);
			return FALSE;
		}
		/* Jetzt ueberpruefen des Headers */
		if (dbHdr.CatMagic != infCatMagic ||
			dbHdr.Version != infVersion ||
			dbHdr.VersionMagic != infVersionMagic)
		{
			if (dbHdr.CatMagic == infCatMagic &&
				dbHdr.Version == oldInfVersion &&
				dbHdr.VersionMagic == oldInfVersionMagic)
			{
				isOldInf = TRUE;
			} else
			{
				MTE_info(Rsc_WrkString(AL_OLDVERSION));
				CatFiles_CloseFile(handle);
				return FALSE;
			}
		} else
		{
			isOldInf = FALSE;
		}
		/* Groesse feststellen */
		CatFiles_Seek(0L, handle, CatFiles_end);
		CatFiles_FilePos(handle, &size);
		CatFiles_Seek(0, handle, CatFiles_start);	/* !!! -> F_File_Size */
		if (isOldInf)
		{
			count = (_UWORD)((size - dbHeaderLength) / sizeof(oldInfEntry));
			oldInfArray = MALLOC(count * sizeof(oldInfEntry), "Infofiles_LoadInfoInf: oldInfArray");
			if (oldInfArray == NULL)
			{
				CatFiles_CloseFile(handle);
				return FALSE;
			}
			/* Neues Array allozieren */
			infArray = MALLOC(count * sizeof(infEntry), "Infofiles_LoadInfoInf: InfArray");
			if (infArray == NULL)
			{
				FREE(oldInfArray, count * sizeof(oldInfEntry));
				CatFiles_CloseFile(handle);
				return FALSE;
			}
			MemSetZero(infArray, count * sizeof(infEntry));

			/* Jetzt Datenblock einlesen */
			CatFiles_Seek(dbHeaderLength, handle, CatFiles_start);
			error = CatFiles_ReadMuch(size - dbHeaderLength, handle, oldInfArray);
			if (error != NoError)
			{
				FREE(infArray, count * sizeof(infEntry));
				FREE(oldInfArray, count * sizeof(oldInfEntry));
				F_ErrorAlert(error);
				CatFiles_CloseFile(handle);
				return FALSE;
			}
			/* Jetzt Daten umkopieren */
			for (i = 0; i < count; i++)
				ConvertInfoV1toV2(&oldInfArray[i], &infArray[i]);
			/* Und altes Array freigeben */
			FREE(oldInfArray, count * sizeof(oldInfEntry));
		} else
		{
			count = (_UWORD)((size- dbHeaderLength) / sizeof(infEntry));
			infArray = MALLOC(count * sizeof(infEntry), "Infofiles_LoadInfoInf: InfArray");
			if (infArray == NULL)
			{
				CatFiles_CloseFile(handle);
				return FALSE;
			}
			/* Jetzt Datenblock einlesen */
			CatFiles_Seek(dbHeaderLength, handle, CatFiles_start);
			error = CatFiles_ReadMuch(size - dbHeaderLength, handle, infArray);
			if (error != NoError)
			{
				F_ErrorAlert(error);
				CatFiles_CloseFile(handle);
				FREE(infArray, count * sizeof(infEntry));
				return FALSE;
			}

			IfNotBigEndian
			{
				for (i = 0, entry = infArray; i < count; i++, entry++)
				{
					entry->i.crc = n2hs(entry->i.crc);
					entry->i.new = n2hs(entry->i.new);
					entry->i.avail = n2hs(entry->i.avail);
					entry->i.isMaus = n2hs(entry->i.isMaus);
					entry->i.ordered = n2hs(entry->i.ordered);
					entry->i.orderIntervall = n2hs(entry->i.orderIntervall);
					entry->i.lastOrdered = n2hl(entry->i.lastOrdered);
					entry->i.selected = n2hs(entry->i.selected);
					entry->i.hasCrc = n2hs(entry->i.hasCrc);
					entry->i.res    = n2hs(entry->i.res);
				}
			}
		}
		CatFiles_CloseFile(handle);
		/* clearNew beachten */
		if (clearNew)
		{
			for (i = 0, entry = infArray; i < count; i++, entry++)
				entry->i.new = FALSE;
		}
		/* Alte Liste loeschen, falls nicht leer */
		Lists_DeleteSimpleList(&infList, mem_free, FALSE);
		/* Jetzt umwandeln in Liste */
		for (i = 0; i < count; i++)
		{
			bSize = INF_ALLOC_SIZE(infArray[i].i.filename);
			entry = MALLOC(bSize, "Infofiles_LoadInfoInf: entry");
			if (entry == NULL)
			{
				FREE(infArray, count * sizeof(infEntry));
				Lists_DeleteSimpleList(&infList, mem_free, FALSE);
				return FALSE;
			}
			/* Anforderungsintervall fuer ITK, ITI und ITG auf 'Immer' setzen,
			 * da nicht mehr aenderbar
			 */
			if (Infofiles_IsNeededInfo(infArray[i].i.mausShort))
			{
				infArray[i].i.ordered = TRUE;
				infArray[i].i.orderIntervall = ORDER_ALWAYS;
			}
			MemCpy(entry, &infArray[i], bSize);
			entry->i.selected = FALSE;
			entry->i.res &= ~ITI_IS_IN_ITI;
			/* Gruppenflags setzen (MH) */
			entry->i.res = (entry->i.res & ~ITI_GROUP_MASK) | Infofiles_GetInfoGroupNumber(entry);

			if (!Lists_AppendEntry(&infList, entry))
			{
				FREE(entry, bSize);
				FREE(infArray, count * sizeof(infEntry));
				Lists_DeleteSimpleList(&infList, mem_free, FALSE);
				return FALSE;
			}
		}
		FREE(infArray, count * sizeof(infEntry));
		/* Jetzt sortieren */
		Infofiles_SortList(&infList);
		/* Und wenn es eine alte Liste war, dann direkt mal sichern */
		if (isOldInf)
		{
			Infofiles_SaveInfoInf();
		}
	} else
	{
		if (dbChanged)
		{
			/* Erstmal Benutzer informieren */
			MTE_info(Rsc_WrkString(AL_NO_INFOINF));
			/* Liste komplett leeren */
			Lists_DeleteSimpleList(&infList, mem_free, FALSE);
		}
		/* Und jetzt Default-Eintraege anlegen. Diese sind:
		 * CAT CATLOG.TXT
		 * ERR PROTOKOL.TXT
		 * LOG Letztes Logfile der Maus
		 * ITI Technische Infofileliste
		 * YGA Technische Gruppenliste alle Gruppen
		 * YGN Technische Gruppenliste neue Gruppen
		 * ITG Technische Gruppenliste
		 * ITK Technische Kommandoliste
		 */
		Infofiles_AddEntry(FALSE, FALSE, FALSE, FALSE, 0, 0, 0, CAT_SHORT_NAME, Rsc_WrkString(INFO_CEINF),        "$D" CATLOG_FILENAME, FALSE, cNoCrc);
		Infofiles_AddEntry(FALSE, FALSE, FALSE, FALSE, 0, 0, 0, ERR_SHORT_NAME, Rsc_WrkString(INFO_CERR),         "$D" CATERR_FILENAME, FALSE, cNoCrc);
		Infofiles_AddEntry(FALSE, FALSE, FALSE, FALSE, 0, 0, 0, LOG_SHORT_NAME, Rsc_WrkString(INFO_LOG),          "$D" LASTLOG_FILENAME, FALSE, cNoCrc);
		Infofiles_AddEntry(TRUE,  TRUE,  FALSE, FALSE, 0, 0, 0, ITI_SHORT_NAME, Rsc_WrkString(INFO_ITI),     ITI_FILENAME, FALSE, cWithCrc);
		if(UserBLK.isInternet)
		{
			Infofiles_AddEntry(TRUE,  TRUE,  FALSE, FALSE, 0, -1, 0, YGB_SHORT_NAME, Rsc_WrkString(INFO_YGB),  YGB_FILENAME, FALSE, cNoCrc);
			Infofiles_AddEntry(TRUE,  TRUE,  FALSE, FALSE, 0, 0, 0, YGN_SHORT_NAME, Rsc_WrkString(INFO_YGN),  YGN_FILENAME, FALSE, cWithCrc);
			Infofiles_AddEntry(TRUE,  FALSE,  FALSE, FALSE, 0, -1, 0, YGA_SHORT_NAME, Rsc_WrkString(INFO_YGA),  YGA_FILENAME, FALSE, cNoCrc);
		}
		else
			Infofiles_AddEntry(TRUE,  TRUE,  FALSE, FALSE, 0, 0, 0, ITG_SHORT_NAME, Rsc_WrkString(INFO_YGA),  ITG_FILENAME, FALSE, cWithCrc);
		Infofiles_AddEntry(TRUE,  TRUE,  FALSE, FALSE, 0, 0, 0, ITK_SHORT_NAME, Rsc_WrkString(INFO_ITK), ITK_FILENAME, FALSE, cWithCrc);
		Infofiles_SaveInfoInf();
		/* Jetzt sortieren */
		Infofiles_SortList(&infList);
		if (clearNew)
		{
			Lists_ResetList(&infList);
			Lists_ScanEntries(&infList, Lists_forward, clearNewProc, NULL);
		}
	}
	return TRUE;
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _BOOL copyEntry(_VOID *e, _VOID *info)
{
	infEntry *entry, *to;

	entry = e;
	to = *((_VOID **)info);
#if !CREATE_DEFAULT
	{
		size_t bSize;

		bSize = INF_ALLOC_SIZE(entry->i.filename);
		MemCpy(to, entry, bSize);
	}
#else
	to->i.crc    = entry->i.crc;
	to->i.new    = FALSE;
	to->i.avail  = FALSE;
	to->i.isMaus = entry->i.isMaus;
	to->i.ordered = maus_short_name_equal(entry->i.mausShort, ITG_SHORT_NAME) ||
		maus_short_name_equal(entry->i.mausShort, ITI_SHORT_NAME);
	to->i.orderIntervall = ORDER_ALWAYS;
	to->i.lastOrdered    = 0;
	to->i.selected       = FALSE;
	to->i.hasCrc         = cNa;
	to->i.res            &= ~ITI_IS_IN_ITI;
	strCpy(to->i.mausShort, entry->i.mausShort);
	strCpy(to->i.mausName, entry->i.mausName);
	strCpy(to->i.filename, entry->i.filename);
#endif

	IfNotBigEndian
	{
		to->i.crc = h2ns(to->i.crc);
		to->i.new = h2ns(to->i.new);
		to->i.avail = h2ns(to->i.avail);
		to->i.isMaus = h2ns(to->i.isMaus);
		to->i.ordered = h2ns(to->i.ordered);
		to->i.orderIntervall = h2ns(to->i.orderIntervall);
		to->i.lastOrdered = h2nl(to->i.lastOrdered);
		to->i.selected = h2ns(to->i.selected);
		to->i.hasCrc = h2ns(to->i.hasCrc);
		to->i.res    = h2ns(to->i.res);
	}
	*((_VOID **)info) = to + 1;
	return FALSE;
}

/*** ---------------------------------------------------------------------- ***/

GLOBAL _BOOL Infofiles_SaveInfoInf(_VOID)
{
	infEntry *infArray;
	_UWORD count;
	FILENO handle;
	infEntry *entry;
	FileHeaderType dbHdr;
	_UBYTE fname[PATH_MAX];
	F_ERROR error;
	
	if (!Infofiles_Init())
		return FALSE;
	if (Lists_ListEmpty(&infList))
		return FALSE;
	
	deleteTitles(&infList);		/* (MH) */
	count = (_UWORD)Lists_NoOfEntries(&infList);
	infArray = MALLOC(count * sizeof(infEntry), "Infofiles_SaveInfoInf: infArray");
	if (infArray == NULL)
	{
		return FALSE;
	}
	MemSetZero(infArray, count * sizeof(infEntry));
	entry = infArray;
	Lists_ResetList(&infList);
	Lists_ScanEntries(&infList, Lists_forward, copyEntry, &entry);
	/* Nun Datei erzeugen */
	strBcpy(fname, DataPath);
#if CREATE_DEFAULT
	F_Path_Append(fname, "info.new");
#else
	F_Path_Append(fname, infoInfFile);
#endif
	handle = CatFiles_CreateFile(fname);
	if (handle < 0)
	{
		F_ErrorAlert(handle);
		FREE(infArray, count * sizeof(infEntry));
		return FALSE;
	}
	dbHdr.CatMagic = h2nl(infCatMagic);
	dbHdr.Version  = h2ns(infVersion);
	dbHdr.VersionMagic = h2ns(infVersionMagic);
	error = CatFiles_WriteMuch(dbHeaderLength, handle, &dbHdr);
	if (error != NoError)
	{
		FREE(infArray, count * sizeof(infEntry));
		F_ErrorAlert(error);
		CatFiles_CloseFile(handle);
		return FALSE;
	}
	error = CatFiles_WriteMuch(count * sizeof(infEntry), handle, infArray);
	FREE(infArray, count * sizeof(infEntry));
	CatFiles_CloseFile(handle);
	if (error != NoError)
	{
		F_ErrorAlert(error);
		return FALSE;
	}
	Protokoll_SendPathUpdate(DataPath);
	return TRUE;
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _VOID ParseFlags(infEntry *entry, CONST _UBYTE *flags)
{
	entry->i.hasCrc = cNa;
	while (*flags != '\0')
	{
		switch (flags[0])
		{
		case 'C':
			switch (flags[1])
			{
			case '+':
				entry->i.hasCrc = cWithCrc;
				break;
			case '-':
				entry->i.hasCrc = cNoCrc;
				break;
			default:
				entry->i.hasCrc = cNa;
				break;
			}
			break;
		}
		if (*++flags != '\0')
			++flags;
	}
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _BOOL searchName(_VOID *e, _VOID *i)
{
	infEntry *entry = e;
	_UBYTE *str = i;

	if (maus_short_name_equal(entry->i.mausShort, str))
		return TRUE;
	return FALSE;
}

/*** ---------------------------------------------------------------------- ***/

/* Formatbeschreibung:

Ansonsten mach' ich hier gleich mal einen Vorschlag fuer das Info-Info, gleich
mal via an die aktuellen Sossen angepasstem Beispiel:

#ITI
:#IIE
::Einfuehrung in die MAUS
:#IIB
::Bedienungsanleitung
:#IIA
::Menue-Kurzanleitung
:#IIG
::Gruppen-Anleitung
:#III
::Technische Informationen
:#IIH
::Hardware der MAUS
:#IIM
::MAUS-Beitrag
:#IIL
::LogIn-Zeiten
:#IIT
::MausTausch-Anleitung
:#IIK
::Kommerzielles
:#INA
::MausNet-Anleitung
:#INK
::Kurze Boxen-Liste
:#INL
::Lange Boxen-Liste
:#ING
::Netzgruppen-Liste
:#INP
::Netzplan
:#IGT
::Boxtitel
:#IGE
::Boxabspann
:#IGS
::Spruch des Tages
:#IGK
::Kurze Gruppenliste
:#IGL
::Lange Gruppenliste
:#ISG                                 | erscheint nur,
::SysOp-Netzgruppenliste              | wenn User
:#ISB                                 | berechtigt
::SysOp-Boxenliste                    | zum Abruf
:#ITG
::Maschinenlesbare Gruppenliste
:#ITI
::Maschinenlesbare Infoliste

Geht das so? Wenn ja, bau' ich das gleich ein ...

MfG Kai
*/

/*** ---------------------------------------------------------------------- ***/

LOCAL _VOID add_info_from_iti(CONST _UBYTE *id, CONST _UBYTE *textScrap, CONST _UBYTE *flagScrap)
{
	infEntry *entry;
	infEntry dummy;
	_UBYTE name[256];

	/* Eintrag suchen und Infos sichern */
	/* Beschreibung zu dem Infofile, also einsetzen */
	Lists_ResetList(&infList);
	if (Lists_ScanEntries(&infList, Lists_forward, searchName, id))
	{
		entry = Lists_CurrentEntry(&infList);
		ParseFlags(entry, flagScrap);
		if (textScrap[0] != '\0')
			strBcpy(entry->i.mausName, textScrap);
		entry->i.res |= ITI_IS_IN_ITI;
	} else
	{
		/* Nicht vorhanden, also neues Infofile anlegen
		 * und anfordern
		 */
		/* Filenamen zusammensetzen */
		strBcpy(name, id);
		strBcat(name, textExt);
		strLwr(name);
		/* Und zu Liste hinzufuegen */
		ParseFlags(&dummy, flagScrap);
		/* Sichern ist nicht noetig, wird am Ende eh gemacht */
		Infofiles_AddEntry(TRUE, FALSE, FALSE, FALSE, 0, -1, 0, id, textScrap, name, FALSE, dummy.i.hasCrc);
	}
}

/*** ---------------------------------------------------------------------- ***/

LOCAL _VOID CheckInfos(_VOID)
{
	_UBYTE idScrap[256];
	_UBYTE textScrap[256];
	_UBYTE flagScrap[256];
	_UBYTE infoScrap[256];
	_UBYTE filename[PATH_MAX];
	TEXTFILENO tf;
	_BOOL infoSet;
	_UBYTE type;
	infEntry *entry;

	strCpy(filename, DataPath);
	F_Path_Append(filename, ITI_FILENAME);
	*infoScrap = '\0';
	*flagScrap = '\0';
	*textScrap = '\0';
	if (mtTextfiles_OpenTextfile(filename, READ, 0, &tf) == GERR_OK)
	{
		infoSet = FALSE;

		Lists_ResetList(&infList);
		while ((entry = Lists_NextEntry(&infList)) != NULL)
			if (entry->i.isMaus)
				entry->i.res &= ~ITI_IS_IN_ITI;
		/* Absolut trivial zu parsen */
		do
		{
			type = mtTextfiles_ReadChar(tf);
			mtTextfiles_ReadLine(tf, idScrap, sizeof(idScrap));
			mtTextfiles_ReadLn(tf);
			switch (type)
			{
			case '#':
				if (infoSet)
				{
					add_info_from_iti(infoScrap, textScrap, flagScrap);
					infoSet = FALSE;
					*infoScrap = '\0';
					*flagScrap = '\0';
					*textScrap = '\0';
				}
				if (idScrap[0] != '\0')
				{
					strCpy(infoScrap, idScrap);
					infoSet = TRUE;
				}
				break;
			case ':':
				strCpy(textScrap, idScrap);
				break;
			case 'F':
				strCpy(flagScrap, idScrap);
				break;
			}
		} while (!mtTextfiles_EndofText(tf));
		mtTextfiles_CloseTextfile(&tf);
		if (infoSet)
		{
			add_info_from_iti(infoScrap, textScrap, flagScrap);
		}
		/* Jetzt alle Maus-Infofiles loeschen, die nicht gerade kamen */
		Lists_ResetList(&infList);
		while ((entry = Lists_NextEntry(&infList)) != NULL)
		{
			if (entry->i.isMaus && !(entry->i.res & ITI_IS_IN_ITI))
			{
				Lists_RemoveEntry(&infList, mem_free);
			}
		}
		/* Jetzt sortieren */
		Infofiles_SortList(&infList);
	}
}

/*** ---------------------------------------------------------------------- ***/

GLOBAL _VOID Infofiles_SetInfoCrc(CONST _UBYTE *inf, CONST _UBYTE *str)
{
	infEntry *entry;
	CRC16 crc;
	_ULONG today;
	_BOOL isNew;
	
	today = ConvertDate_DateInNum(Tgetdate());
	isNew = FALSE;
	Lists_ResetList(&infList);
	while ((entry = Lists_NextEntry(&infList)) != NULL)
	{
		if (entry->i.isMaus && maus_short_name_equal(entry->i.mausShort, inf))
		{
			crc = StrToNum(str, 10, NULL, NULL);
			if (Infofiles_GetInfoGroupNumber(entry) == ITI_INFOFILE_O)	/* (MH) */
				entry->i.new = TRUE;
			else
				entry->i.new = !entry->i.avail || crc != entry->i.crc || inf[0] == 'J';
			isNew = entry->i.new;
			entry->i.avail = TRUE;
			entry->i.crc = crc;
			entry->i.lastOrdered = today;
			if (entry->i.orderIntervall == ORDER_ONCE)
			{
				entry->i.orderIntervall = ORDER_ALWAYS;
				entry->i.ordered = FALSE;
			}
			break;
		}
	}
	if (isNew)
	{
		if (maus_short_name_equal(inf, ITI_SHORT_NAME))
		{
			/* Maschinenlesbare Infofileliste ist neu!
			 * Also Abgleich zwischen Infoliste und internen Daten
			 */
			CheckInfos();
		}
	}
}

/*** ---------------------------------------------------------------------- ***/

/* Liefert einen Infostring zurueck. Bei Mode == IFIRST wird das erste Infofile
 * zurueckgeliefert, bei Mode == INEXT wird der naechste zurueckgeliefert.
 * Wenn alle Infostrings geliefert wurden, dann wird FALSE zurueckgegeben
 * (kann auch bei IFIRST passieren)
 * Zwischen den Aufrufen sollte nicht anders auf die Liste
 * zugegriffen werden, da ansonsten evtl. nicht alle angefordert
 * werden oder einige doppelt angefordert werden.
 */
_BOOL Infofiles_GetInfoCrc(_UBYTE *str, infMode mode)
{
	_ULONG today;
	Date dt;
	Time ti;
	tWeekday akt_wday;
	_UWORD stunde;
	infEntry *entry;
	_UBYTE numstr[10];

	*str = '\0';
	today = ConvertDate_DateInNum(Tgetdate());
	dt = ConvertDate_UnpackDate(Tgetdate());
	ti = ConvertDate_UnpackTime(Tgettime());
	akt_wday = ConvertDate_WeekDay(&dt);
	stunde = ti.hour;
	if (mode == IFIRST)
		Lists_ResetList(&infList);
	while ((entry = Lists_NextEntry(&infList)) != NULL)
	{
		/* Infofile auf jeden Fall neu anfordern (MH) */
		if (entry->i.res & ITI_ORDERED_NOW)
		{
			entry->i.ordered = TRUE;
			entry->i.avail = FALSE;
		}
		if (entry->i.isMaus && entry->i.ordered)
		{
			if (entry->i.avail)
			{
				/* Bestellung nach Wochentag (MH) */
				if (entry->i.res & ITI_ORDERED_WDAY)	
				{
					if (entry->i.orderIntervall == akt_wday && 
					    entry->i.lastOrdered < today)	/* einmal taeglich */
					{
						CardToStr(entry->i.crc, 0, numstr, sizeof(numstr));
					} else
					{
						continue;
					}
				} else
				{
					if ((entry->i.orderIntervall <= 0) ||
						((entry->i.lastOrdered + entry->i.orderIntervall == today)
							&& stunde >= 3 /* Neues Infofile erst nach dem Nachtnetz anfordern */ ) ||
						(entry->i.lastOrdered + entry->i.orderIntervall < today)  || /* ausser es ist schon noch etwas aelter */
						(entry->i.lastOrdered > today))
					{
						CardToStr(entry->i.crc, 0, numstr, sizeof(numstr));
					} else
					{
						continue;
					}
				}
			} else
			{
				IntToStr(-1, 0, numstr, sizeof(numstr));
			}
			strCpy(str, entry->i.mausShort);
			if (entry->i.hasCrc == cWithCrc || (entry->i.hasCrc == cNa && str[0] != 'J'))
			{
				strCat(str, " ");
				strCat(str, numstr);
			}
			return TRUE;
		}
	}
	return FALSE;
}

/*** ---------------------------------------------------------------------- ***/

GLOBAL _BOOL Infofiles_AddEntry(
	_BOOL isMausInf,
	_BOOL order,
	_BOOL avail,
	_BOOL new,
	CRC16 crc,
	_WORD inter,
	_ULONG lastOrder,
	CONST _UBYTE *shortname,
	CONST _UBYTE *name,
	CONST _UBYTE *file,
	_BOOL saveInf,
	tCrcState withCrc)
{
	size_t bSize;
	infEntry *entry;

	if (isMausInf)
	{
		/* Immer I??.TXT als Filename! */
		bSize = INF_ALLOC_SIZE(shortname) + (sizeof(textExt) - 1);
	} else
	{
		bSize = INF_ALLOC_SIZE(file);
	}
	entry = MALLOC(bSize, "Infofiles_AddEntry");
	if (entry == NULL)
	{
		return FALSE;
	}
	MemSetZero(entry, bSize);
	entry->i.crc = crc;
	entry->i.avail = avail;
	entry->i.new   = new;
	entry->i.isMaus = isMausInf;
	entry->i.ordered = order;
	entry->i.orderIntervall = inter;
	entry->i.lastOrdered = lastOrder;
	entry->i.hasCrc = withCrc;
	entry->i.res |= ITI_IS_IN_ITI;
	strBcpy(entry->i.mausShort, shortname);
	strBcpy(entry->i.mausName, name);
	if (!isMausInf)
	{
		strCpy(entry->i.filename, file);
	} else
	{
		strCpy(entry->i.filename, shortname);
		strCat(entry->i.filename, textExt);
		strLwr(entry->i.filename);
	}
	if (!Lists_AppendEntry(&infList, entry))
	{
		FREE(entry, bSize);
		return FALSE;
	}
	if (saveInf)
		Infofiles_SaveInfoInf();
	return TRUE;
}

/*** ---------------------------------------------------------------------- ***/

/* Entfernt einen Eintrag (MH) */
GLOBAL _BOOL Infofiles_DeleteEntry(CONST _UBYTE *shortname)
{
	infEntry *info;

	Lists_ResetList(&infList);
	if (Lists_ScanEntries(&infList, Lists_forward, searchName, shortname))
	{
		info = Lists_CurrentEntry(&infList);
		if (info != NULL && Lists_RemoveEntry(&infList, mem_free))
			return TRUE;
	}
	return FALSE;
}	

/*** ---------------------------------------------------------------------- ***/

/* Aktualisiert die Liste der OUT-Bloecke (MH) */

GLOBAL _VOID Infofiles_CleanUpOutBlocks(_VOID)
{
	TEXTFILENO tf;
	_UBYTE fname[PATH_MAX];
	_UBYTE tScrap[256];
	_BOOL exists;
	infEntry *info;
	
	Lists_ResetList(&infList);
	while ((info = Lists_NextEntry(&infList)) != NULL)
	{
		if (Infofiles_GetInfoGroupNumber(info) == ITI_INFOFILE_O)
		{
			exists = (Infofiles_BuildName(info, fname)!=0);
			/* neuer OUT-Block? */
			if (info->i.new)
			{
				if (exists && mtTextfiles_OpenTextfile(fname, READ, 0, &tf) == GERR_OK)
				{
					/* Erste Zeile aus OUT-Block als Beschreibung kopieren */
					mtTextfiles_ReadLine(tf, tScrap, sizeof(tScrap));
					if (tScrap[0] == '!')
						strMcpy(info->i.mausName, sizeof(info->i.mausName), tScrap + 1);
					else
						strMcpy(info->i.mausName, sizeof(info->i.mausName), DEFAULT_OUTBLOCK);
					mtTextfiles_CloseTextfile(&tf);
				} else
				{
					/* Doch nicht gefunden, Eintrag loeschen */
					info->i.avail = FALSE;
				}
			} else
			{
				if (!exists)
				{
					Lists_RemoveEntry(&infList, mem_free);
				}
			}
		}
	}
}
	
/*** ---------------------------------------------------------------------- ***/

/* Setzt das ITI_ORDERED_NOW-Flag zurueck (MH) */
GLOBAL _VOID Infofiles_ResetOrderedNowFlag(_VOID)
{
	infEntry *info;
	
	Lists_ResetList(&infList);
	while ((info = Lists_NextEntry(&infList)) != NULL)
	{
		info->i.res &= ~ITI_ORDERED_NOW;
	}
}
	
/*** ---------------------------------------------------------------------- ***/


GLOBAL _ULONG Infofiles_BuildName(infEntry *info, _UBYTE *fname)
{
	_ULONG ret=0l;
	_UBYTE filename[PATH_MAX];
	_BOOL exists;

	if (info->i.isMaus)
	{
		strCpy(fname, DataPath);
		strCpy(filename, info->i.mausShort);
		strCat(filename, textExt);
		strLwr(filename);
		F_Path_Append(fname, filename);
	} else
	{
		strCpy(filename, DataPath);
		F_Path_Append(filename, "");
		strCpy(fname, info->i.filename);
		CatGlobal_ReplaceStr(fname, PATH_MAX, "$D", filename);
	}
	/* Jetzt Datei auf Existenz pruefen und auf Laenge != 0 */
	if ((ret=CatFiles_FileSize(fname, &exists)) == 0)
	{
		/* Erstmal Listeintrag updaten, Infofile nicht da! */
		if (!exists)
		{
			info->i.avail = FALSE;
			info->i.new = FALSE;
		}
		ret=0l;
	}
	return ret;
}

/*** ---------------------------------------------------------------------- ***/

/* Liefert zum Kuerzel des Infofiles den Filenamen.
 * FALSE: Infofile nicht vorhanden
 */

GLOBAL _ULONG Infofiles_GetInfoFilename(CONST _UBYTE *shortname, _UBYTE *fname)
{
	_ULONG ret=0l;
	infEntry *info;

	Lists_ResetList(&infList);
	if (Lists_ScanEntries(&infList, Lists_forward, searchName, shortname))
	{
		info = Lists_CurrentEntry(&infList);
		if (info != NULL && (ret=Infofiles_BuildName(info, fname))!=0l)
			return ret;
	}
	return 0l;
}

/*** ---------------------------------------------------------------------- ***/

/* Liefert zurueck, ob ein Kuerzel ein Infofile ist oder nicht
 */
GLOBAL _BOOL Infofiles_IsInfoFile(CONST _UBYTE *shortname, _BOOL onlyMaus)
{
	infEntry *info;
	_BOOL found;

	Lists_ResetList(&infList);
	found = Lists_ScanEntries(&infList, Lists_forward, searchName, shortname);
	if (found && onlyMaus)
	{
		info = Lists_CurrentEntry(&infList);
		if (info != NULL)
			found = info->i.isMaus;
	}
	return found;
}

/*** ---------------------------------------------------------------------- ***/

/* Liefert zurueck, ob ein Infofile verfuegbar ist
 */
GLOBAL _BOOL Infofiles_Available(CONST _UBYTE *shortname, _BOOL onlyMaus)
{
	infEntry *info;
	_BOOL found;

	Lists_ResetList(&infList);
	found = Lists_ScanEntries(&infList, Lists_forward, searchName, shortname);
	if (found)
	{
		info = Lists_CurrentEntry(&infList);
		if (info != NULL)
		{
			if (onlyMaus)
				found = info->i.isMaus;
			if (!info->i.avail)
				found = FALSE;
		}
	}
	return found;
}

/*** ---------------------------------------------------------------------- ***/

GLOBAL _VOID Infofiles_Exit(_VOID)
{
	if (infList.root != NULL)
		Lists_DeleteSimpleList(&infList, mem_free, TRUE);
}

