#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <tos.h>

#define MAXFUNCS 5000
#define NAMESIZE 8

typedef struct {
	char name[NAMESIZE];
	long val;
	long size;
} FUNC;

FUNC extfuncs[MAXFUNCS+1];

#include "funcs.c"


typedef struct {
	char name[NAMESIZE];
	short type;
	long val;
} SYMBOL;



#define OP_LINK 0x4e56 /* link a6,# */
#define OP_UNLK 0x4e5e /* unlk a6 */
#define OP_RTS  0x4e75 /* rts */
#define OP_TRAP 0x4e40

#define MAGIC 0x50524F46l


static int verbose = 0;
static int quiet = 0;


#ifdef __STDC__
# define _(x) x
#else
# define _(x) ()
#endif

#define FALSE 0
#define TRUE  1




static void add_extension(char *filename, char *ext)
{
    char *p;
    
    if ((p = strrchr(filename, '\\')) == NULL) p = filename;
    else ++p;
    if ((strchr(p, '.')) == NULL)
    {
        strcat(filename, ".");
        strcat(filename, ext);
    }
}


static void usage(void)
{
}


static void scan_option(char *argp)
{
	while (*++argp)
	{
		switch (tolower(*argp))
		{
		case 'v':
			verbose++;
			break;
		case 'q':
			quiet++;
			break;
		case '?':
			usage();
			exit(1);
		default:
			fprintf(stderr, "unknown option '%c', ignored\n", *argp);
			break;
		}
	}
}


static int comp(const void *str1, const void *str2)
{
	const char *s1 = (const char *)str1;
	const char *s2 = (const char *)str2;
	int n = NAMESIZE+1;
	
	while (--n && *s1)
	{
		if (*s1 != *s2)
			return(*s1 - *s2);
		s1++;
		s2++;
	}
	if (n == 0 || *s2 == ' ')
		return 0;
	return -1;
}


static int fcomp(const void *f1, const void *f2)
{
	if (((FUNC *)f1)->val < ((FUNC *)f2)->val)
		return(-1);
	return(1);
}


static int replace(FILE *fp, FUNC *f)
{
	short w;
	long pos;
	long end;
	char name[NAMESIZE+1];
	
	memcpy(name, f->name, NAMESIZE);
	name[NAMESIZE] = '\0';
	fseek(fp, f->val+sizeof(PRGHDR), SEEK_SET);
	fread(&w, 2, 1, fp);
	if (quiet)
	{
		putchar('.');
		fflush(stdout);
	}
	if (w != OP_LINK)
	{
		if (w == OP_TRAP)
		{
			if (!quiet)
				fprintf(stderr, "function %s already fixed\n", name);
		} else
		{
			if (quiet)
				putchar('\n');
			fprintf(stderr,	"function %s has no LINK\n", name);
		}
		return FALSE;
	}
	if (!quiet)
		printf("fixing %s ", name);
	fseek(fp, -2l, SEEK_CUR);
	w = OP_TRAP;
	fwrite(&w, 2, 1, fp);
	fseek(fp, 0l, SEEK_CUR);
	pos = ftell(fp) - sizeof(PRGHDR);
	end = pos + f->size;
	while (pos < end)
	{
		fread(&w, 2, 1, fp);
		if (w == OP_UNLK)
		{
			fread(&w, 2, 1, fp);
			if (w == OP_RTS)
			{
				fseek(fp, -4l, SEEK_CUR);
				w = OP_TRAP;
				fwrite(&w, 2, 1, fp);
				if (!quiet)
				{
					putchar('.');
					fflush(stdout);
				}
				fseek(fp, 2l, SEEK_CUR);
			}
			pos += 2;
		}
		pos += 2;
	}
	if (!quiet)
		printf("\n");
	return(TRUE);
}


#define _WORD short


main(int argc, char **argv)
{
	char *argp;
	char *filename;
	FILE *fp;
	PRGHDR hdr;
	long size;
	SYMBOL sym;
	char name[NAMESIZE+1];
	_WORD i, n;
	_WORD numfuncs;
	char fnamebuf[100];
	long magic;
	
	filename = NULL;
	while (--argc)
	{
		if (*(argp = *++argv) == '-')
		{
			scan_option(argp);
		} else
		{
			if (filename != NULL)
			{
				usage();
				return 1;
			}
			filename = argp;
		}
	}
	if (filename == NULL)
	{
		usage();
		return 1;
	}
	strcpy(fnamebuf, filename);
	if ((fp = fopen(fnamebuf, "r+b")) == NULL)
	{
		add_extension(fnamebuf, "ttp");
		if ((fp = fopen(fnamebuf, "r+b")) == NULL)
		{
			fprintf(stderr, "can't open '%s'\n", fnamebuf);
			return(1);
		}
	}
	if (fread(&hdr, 1, sizeof(hdr), fp) != sizeof(hdr) ||
		hdr.p_magic != 0x601a)
	{
		fprintf(stderr, "not executable: %s\n", fnamebuf);
		return(1);
	}
	if ((size = hdr.p_slen) == 0)
	{
		fprintf(stderr, "no symbol table\n");
		return(1);
	}
	fseek(fp, 12l, SEEK_CUR);
	fread(&magic, 1, 4, fp);
	if (magic != MAGIC)
	{
		fprintf(stderr, "wrong startup code: %s\n", fnamebuf);
		return(1);
	}
	fseek(fp, hdr.p_tlen+hdr.p_dlen+sizeof(hdr), SEEK_SET);
	numfuncs = 0;
	while (size > 0)
	{
		fread(&sym, sizeof(SYMBOL), 1, fp);
		strncpy(name, sym.name, NAMESIZE);
		name[NAMESIZE] = 0;
		if (verbose)
			printf("%-*.*s $%08lX %04X\n",
				NAMESIZE, NAMESIZE, name,
				sym.val,
				sym.type);
		if ((sym.type & 0x8300) == 0x8200)
		{
			if (numfuncs == MAXFUNCS)
			{
				fprintf(stderr, "table overflow\n");
				return(1);
			}
			memcpy(extfuncs[numfuncs].name, name, NAMESIZE);
			extfuncs[numfuncs].val = sym.val;
			numfuncs++;
		}
		size -= sizeof(SYMBOL);
	}
	qsort(extfuncs, (size_t)numfuncs, sizeof(FUNC), fcomp);
	qsort(funclist, nfuncs, NAMESIZE, comp);
	extfuncs[numfuncs+1].val = hdr.p_tlen;
	n = 0;
	for (i = 0; i < numfuncs; i++)
	{
		extfuncs[i].size = extfuncs[i+1].val - extfuncs[i].val;
		if (bsearch(extfuncs[i].name, funclist, nfuncs, NAMESIZE, comp) == NULL)
		{
			extfuncs[n++] = extfuncs[i];
		}
	}
	numfuncs = n;
	for (i = 0; i < numfuncs; i++)
	{
		replace(fp, &extfuncs[i]);
	}
	if (quiet)
		putchar('\n');
	fclose(fp);
	return 0;
}
