/*
 * Diese Datei ist Teil des Megamax Modula-2 Entwicklungssytems,
 * und darf in dieser Form nur als Teil des CAT-Projekts
 * verwendet werden.
 * 
 * CAT ist Freeware; es darf unter den Bedingungen der
 * CAT Public License weitergegeben und/oder modifiziert
 * werden.
 *
 * Eine Kopie der CAT Public License liegt diesem Paket bei
 * (siehe CATPL.TXT).
 *
 * Megamax Modula-2 ist Copyright (c) by Thomas Tempelmann.
 *
 * Jegliche Veraenderung dieses Hinweises ist NICHT ERLAUBT.
 */

/*
 * Die vorliegende Version wurde nach C portiert von
 * Thorsten Otto (Thorsten_Otto@do.maus.ruhr.de)
 */


#ifndef __MOS_LISTS_H__
#define __MOS_LISTS_H__

#ifndef __PORTAB_H__
#include <portab.h>
#endif
#ifndef __RO_MEM_H__
#include <ro_mem.h>
#endif

/*
 * Allgemeine Listenverwaltung.
 *
 * Nach 'ADTLists' aus: Dal Cin, Lutz, Risse: Programmierung in Modula-2.
 *
 * Erstellt 22.3.87, TT
 */

/* ----------------------------------------------------------------------------

    Listen allgemein erlauben, mehrere Datenfelder nacheinander in belie-
    biger Anzahl anzulegen und selektiv wieder zu entfernen. Im Gegensatz
    dazu besteht eine Tabelle aus einem einmalig anfangs zu definierenden
    Speicherbereich, in dem die Daten abgelegt werden und das Ein- oder
    Ausfuegen nur durch aufwendiges Umkopieren moeglich ist. Bei Listen wird
    immer nur soviel Speicher belegt, wie Datenfelder benoetigt werden.

    Die Verwaltung von Listen ist im Vergleich zu Tabellen aufwendiger.
    Dieses Modul bietet dafuer komfortable Funktionen.

    Eine Liste wird ueber eine Variable der Type 'List' angelegt und ver-
    waltet. Ueber sie werden alle Elemente der Liste verkettet. Ist die
    Liste leer, existiert nur das Wurzel-Element, das zur internen Ver-
    waltung dient. Auf die Wurzel kann das Anwenderprogramm nicht zugrei-
    fen. Wird die Liste um neue Elemente erweitert, sind alle Elemente
    ringweise verkettet. Gestartet wird ueblicherweise bei der Wurzel.
    Von dort aus kann jedes Element schrittweise im Kreis erreicht werden.
    Am Ende wird wieder die Wurzel erreicht. Da die Liste intern mit
    Zeigern verkettet wird, ist 'List' als ein Record mit zwei Zeigern
    auf eine interne Datenstruktur definiert. Das Record-Element
    'List.root' zeigt immer auf die Wurzel der Liste, 'List.current' zeigt
    auf das gerade bearbeitete Element der Liste. 'List.user' kann vom
    Anwenderprogramm verwendet werden, um 'current' zeitweise darin zu
    sichern und rueckzuspeichern. 'List.user' wird von diesen Funktionen
    nur Anfangs beim Erzeugen einer Liste mit 'CreateList' auf 'root'
    gesetzt, danach nicht mehr veraendert, es sei denn, beim Loeschen eines
    Elements ('RemoveEntry') war 'current' = 'user'. Dann wird 'user', wie
    'current', auf den Vorgaenger gesetzt.

    Da Modula-2 keine generischen Datenstukturen kennt, mu, um beliebige
    Daten als Liste verwalten zu koennen, das Anwenderprogramm die Daten
    selbst anlegen (z.B. mit NEW oder ALLOCATE). Die Listenfunktionen
    speichern nur die Adressen (Zeiger auf) die Daten. Bei Entfernen
    eines Elements mu daher nicht nur das Element aus der Liste wieder
    entfernt werden (mit einer Funktion dieses Moduls) sondern auch der
    Speicher fuer das Datum wieder freigegeben werden (z.B. mit DISPOSE
    oder DEALLOCATE).

    Hinweis
    -------

    In einigen Modulen des MOS werden Listen verwendet (z.B. in 'Paths',
    'Loader'). Dabei werden diese Listen in der Regel durch einen
    neuen TYPE definiert, welcher wiederum exportiert wird (z.B in
    'Paths':  "TYPE  PathList = List (* OF PathEntry *)"). Hinter
    den Definitionen erscheint dann in Klammern "OF" mit einer anderen
    Type. Dies bedeutet, dass alle Listenelemente ueber diese Type
    angesprochen werden muessen. So sind beispielsweise Funktionsergebnisse,
    wie 'NextEntry' (Lists-Funktion, s.u.), die ja normalerweise einen
    beliebigen ADDRESS-Typen darstellen, hier als 'PathEntry' zu verwenden
    ('PathEntry' ist beispielsweise ein "POINTER TO PathStr"), damit
    der korrekte Zugriff auf die Listenelemente gewaehrleistet ist.


    Beispiel
    --------

    Ein Liste wird erzeugt und darin drei Daten eingefuegt, ausgegeben,
    dann ein Element geloescht und wieder alle Elemente ausgegeben.
    Am Ende wird die gesamte Liste entfernt.


    MODULE ListDemo;

    IMPORT InOut, Lists, Strings;

    FROM SYSTEM IMPORT ADDRESS, ADR;

    FROM Storage IMPORT ALLOCATE, DEALLOCATE;

    TYPE DataStr = ARRAY [0..20] OF _UBYTE;
         DatenType = RECORD
                       (* beliebige Datendefinition *)
                       no: CARDINAL;
                       s: DataStr
                     END;

         Element = POINTER TO DatenType;

    VAR myList: Lists.List;
        error: _BOOL;

    PROCEDURE stop (s: ARRAY OF _UBYTE);               (* Fehleranzeige *)
      VAR c: _UBYTE;
      BEGIN InOut.WriteString (s); InOut.Read (c); InOut.WriteLn END stop;

    PROCEDURE insElem (no: CARDINAL; s: ARRAY OF _UBYTE); (* Elem. anfuegen *)
      VAR elem: Element; ok: _BOOL;
      BEGIN
        NEW (elem);                                  (* Speicher anfordern *)
        elem^.no:= no;                               (* Daten eintragen *)
        Strings.Assign (s, elem^.s, ok);
        Lists.InsertEntry (myList, elem, error); (* Elem. in Liste eintragen *)
        IF error THEN stop ('Kein Speicher') END
      END insElem;

    PROCEDURE showElems;                           (* Alle Elemente anzeigen *)
      VAR elem: Element;
      BEGIN
        Lists.ResetList (myList);               (* Liste auf Anfang (Wurzel) *)
        InOut.WriteString ('Alle Daten:');
        InOut.WriteLn;
        REPEAT
          elem:= Lists.NextEntry (myList);           (* Naechstes Elem. holen *)
          IF elem <> NIL THEN                        (* Wenn nicht Wurzel... *)
            InOut.WriteString ('no: ');
            InOut.WriteCard (elem^.no, 0);
            InOut.WriteString (', s: ');
            InOut.WriteString (elem^.s);             (* ... dann ausgeben    *)
            InOut.WriteLn
          END
        UNTIL elem = NIL  (* oder: "UNTIL List.LastEntry (myList)" *)
      END showElems;

    PROCEDURE checkElem (elem, txt: ADDRESS): _BOOL;
      VAR elem0: Element; txt0: POINTER TO DataStr;
      BEGIN                                               (* Elemente pruefen *)
        elem0:= elem; txt0:= txt;
        RETURN Strings.StrEqual (elem0^.s, txt0^)
                                        (* Vergl. aktuelles 's' mit Suchtext *)
      END checkElem;

    PROCEDURE search (txt: DataStr);             (* Sucht bestimmtes Element *)
      VAR found: _BOOL;
      BEGIN
        Lists.ScanEntries (myList, Lists.forward, checkElem, ADR (txt), found);
          (* Hiermit wird, wenn 'checkElem' fuendig wird, 'current' auf *)
          (* das gefundene Element gesetzt.                            *)
        IF NOT found THEN stop ('Nicht gefunden.') END
      END search;

    PROCEDURE removeElem;
      VAR elem: Element;
      BEGIN
        elem:= Lists.CurrentEntry (myList); (* Akt. Element ermitteln *)
        DISPOSE (elem);                     (* Speicher v. Element freigeben *)
        Lists.RemoveEntry (myList, error);  (* Element aus Liste entfernen *)
      END removeElem;

    PROCEDURE removeAllElems;                   (* Alle Elemente freigeben *)
      BEGIN
        Lists.ResetList (myList);
        WHILE Lists.NextEntry (myList) <> NIL DO removeElem END;
      END removeAllElems;

    VAR c: _UBYTE;

    BEGIN
      Lists.CreateList (myList, error);               (* Liste anlegen *)
      IF error THEN stop ('Kein Speicher') END;
      insElem (1, 'Eins');                            (* Daten anfuegen *)
      insElem (2, 'Zwei');
      insElem (3, 'Drei');
      showElems;                                  (* Alle Elemente zeigen *)
      search ('Zwei');                            (* Elem. m. 'Zwei' suchen *)
      removeElem;                                 (* Dies Element entfernen *)
      showElems;                                  (* Alle Elemente zeigen *)
      removeAllElems;                             (* Alle Elemente entfernen *)
      Lists.DeleteList (myList, error);           (* Liste entfernen *)
      IF error THEN stop ('Liste nicht leer !') END;
      InOut.Read (c)
    END ListDemo.

---------------------------------------------------------------------------- */

#ifdef __MOS_LISTS_IMPLEMENTATION__
struct _lcarrier;
#else
struct _lcarrier { int dummy; };
#endif
typedef struct _lcarrier *LCarrier;

typedef struct List {
	LCarrier current;
	LCarrier root;
	_VOID *user;
} List;

/* entry: */
/* info : */
typedef _BOOL (*LCondProc)(_VOID *entry, _VOID *info);

/* vorwaerts  */
/* rueckwaerts */
typedef enum {
	Lists_forward, Lists_backward
} LDir;


_VOID Lists_InitList(List *l);
/*
 * Initialisiert die Liste 'l', sodass sie definierte Werte
 * enthaelt (hierbei wird nicht, wie bei CreateList, eine Liste
 * angelegt, lediglich wird ein Grundzustand hergestellt)
 * Diese Funktion ist nur notwendig, wenn schon auf die Liste
 * mit anderen Lists-Funktionen zugegriffen werden kann, bevor
 * sie mit 'CreateList' angelegt wird.
 */

#if DEBUG_ALLOC
_BOOL Lists_CreateList(List *list, CONST _UBYTE *from);
#define Lists_CreateList(list, from) Lists_CreateList(list, from)
#else
_BOOL Lists_CreateList(List *list);
#define Lists_CreateList(list, from) Lists_CreateList(list)
#endif
/*
 * Erzeugt neue, leere Liste.
 * 'error':= "Kein Speicher mehr frei"
 */

_BOOL Lists_DeleteList(List *list);
/*
 * Entfernt leere Liste.
 * 'error':= "Liste ist nicht leer"
 */

_VOID Lists_ResetList(List *list);
/*
 * Ernennt die Wurzel (list.root) zum aktuellen Eintrag.
 */

_BOOL Lists_ListEmpty(CONST List *list);
/*
 * Liefert TRUE, wenn die Liste leer ist, also keine Eintraege hat.
 */

_BOOL Lists_InsertEntry(List *list, CONST _VOID *entry);
/*
 * Fuegt Eintrag hinter an aktueller Position ein; 'list.current' wird auf
 * neuen Eintrag gesetzt.
 * 'error':= "Kein Speicher mehr frei"
 */
typedef _VOID (*lists_delete_proc)(_VOID *entry);

_BOOL Lists_RemoveEntry(List *list, lists_delete_proc delete);
/*
 * Entfernt aktuellen Eintrag. 'current' wird auf Vorgaenger gesetzt.
 * 'error':= "list.current zeigt auf Wurzel (=list.root)"
 */

_VOID Lists_DeleteSimpleList(List *l, lists_delete_proc delete, _BOOL killCarrier);

_BOOL Lists_AppendEntry(List *list, CONST _VOID *entry);
/*
 * Fuegt Eintrag am Listenende an; 'list.current' bleibt unveraendert.
 * 'error':= "Kein Speicher mehr frei"
 */

_VOID *Lists_CurrentEntry(List *list);
/*
 * Liefert aktuellen Eintrag; liefert NIL, wenn 'current'='root'.
 */

LCarrier Lists_GetCurrent(List *list);
_VOID Lists_SetCurrent(List *list, LCarrier current);

_VOID *Lists_NextEntry(List *list);
/*
 * Liefert naechsten Eintrag, setzt 'current' darauf; liefert NIL, wenn
 * 'current'='root'.
 */

_VOID Lists_SetNextEntry(List *list,void* data);
/*
 * Setzt den naechsten Eintrag, sinnvoll nach Sortieren, so braucht die
 *	Liste nicht gelscht und neu aufgebaut zu werden.
 */

_VOID *Lists_PrevEntry(List *list);
/*
 * Liefert vorigen Eintrag, setzt 'current' darauf; liefert NIL, wenn
 * 'current'='root'.
 */

_BOOL Lists_FindEntry(List *list, CONST _VOID *entry);
/*
 * Durchsucht ganze Liste nach einem Eintrag in Vorwaertsrichtung.
 * Faengt dabei bei der Wurzel an.
 * Wird der Eintrag gefunden, zeigt 'list.current' darauf. Sonst wird
 * 'list.current' auf die Wurzel gesetzt (wie bei ResetList).
 * 'found':= "Eintrag gefunden"
 */

_BOOL Lists_ScanEntries(List *list, LDir dir, LCondProc cond, CONST _VOID *info);
/*
 * Durchsucht die Liste in der gewuenschten Reihenfolge ('dir'), beginnend
 * beim naechsten Eintrag (wie 'NextEntry').
 * Dabei wird jedesmal die Prozedur 'cond' aufgerufen und ihr der aktuelle
 * Eintrag uebergeben. Diese Prozedur kann dann entscheiden, ob weitergesucht
 * werden soll, indem sie FALSE (weiter) oder TRUE (gefunden, abbrechen)
 * zurueckgibt. Ihr kann, z.B. als Zeiger auf einen Vergleichswert, in 'info'
 * ein Pointer uebergeben werden.
 * Die an 'cond' uebergebene Procedure darf lokal sein!
 * 'found':= "Eintrag gefunden ('cond' lieferte TRUE)"
 */

_BOOL Lists_EndOfList(List *list);
/*
 * Liefert TRUE, wenn 'current' = 'root' ist, also wenn der aktuelle
 * Eintrag die Wurzel ist.
 */

_ULONG Lists_NoOfEntries(List *list);
/*
 * Liefert Anzahl der vorhandenen Eintraege.
 */

_BOOL Lists_FirstEntry(List *list);
/*
 * Liefert TRUE, wenn der Vorgaenger die Wurzel ist.
 */

_BOOL Lists_LastEntry(List *list);
/*
 * Liefert TRUE, wenn der Nachfolger die Wurzel ist.
 */

#endif /*__MOS_LISTS_H__*/
